#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import geoclide as gc
import pytest
from geoclide.shapes import get_intersect_dataset

Q1 = [gc.Disk(radius=3., inner_radius=1.2),
      gc.Sphere(radius=3.),
      gc.Spheroid(radius_xy=3., radius_z=2.),
      gc.Spheroid(radius_xy=2., radius_z=3.)]


def test_sphere():
    # Take README exemple
    vza = 45.
    vaa = 45.
    sat_altitude = 700.
    origin = gc.Point(0., 0., 0.)
    theta = vza
    phi = -vaa

    dir_to_sat = gc.Vector(0., 0., 1.)
    dir_to_sat = gc.get_rotateY_tf(theta)(dir_to_sat)
    dir_to_sat = gc.get_rotateZ_tf(phi)(dir_to_sat)
    ray = gc.Ray(o=origin, d=dir_to_sat)

    earth_radius = 6378. 
    oTw = gc.get_translate_tf(gc.Vector(0., 0., -earth_radius))
    sphere_sat_alti = gc.Sphere(radius=earth_radius+sat_altitude, oTw=oTw)  # apply oTw to move the sphere center to earth center
    ds_sp = gc.calc_intersection(sphere_sat_alti, ray)

    p = ds_sp['phit'].values

    assert (np.isclose(p[0], 472.61058011386376, 0., 1e-15))
    assert (np.isclose(p[1], -472.61058011386365, 0., 1e-15))
    assert (np.isclose(p[2], 668.3722921180423, 0., 1e-15))
    assert (ds_sp['is_intersection'] == sphere_sat_alti.is_intersection(ray))


def test_spheroid():
    oblate = gc.Spheroid(radius_xy=3., radius_z=1.5)
    prolate = gc.Spheroid(radius_xy=1.5, radius_z=3.)
    r1 = gc.Ray(o=gc.Point(2.5, 0., 10.), d=(gc.Vector(0., 0., -1.)))
    r2 = gc.Ray(o=gc.Point(10., 0., 2.5), d=(gc.Vector(-1., 0., 0.)))
    
    ds = gc.calc_intersection(oblate, r1)
    p = ds['phit'].values
    assert (p[0] == 2.5)
    assert (p[1] == 0.)
    assert (np.isclose(p[2], 0.8291561975888655, 0., 1e-15))
    n = ds['nhit'].values
    assert (np.isclose(n[0], 0.6019292654288356, 0., 1e-15))
    assert (n[1] == 0.)
    assert (np.isclose(n[2], 0.7985494095046983, 0., 1e-15))
    assert (ds['is_intersection'].values == oblate.is_intersection(r1))

    ds = gc.calc_intersection(prolate, r2)
    p = ds['phit'].values
    assert (np.isclose(p[0], 0.8291561975888655, 0., 1e-15))
    assert (p[1] == 0.)
    assert (p[2] == 2.5)
    n = ds['nhit'].values
    assert (np.isclose(n[0], 0.7985494095046906, 0., 1e-15))
    assert (n[1] == 0.)
    assert (np.isclose(n[2], 0.6019292654288461, 0., 1e-15))

    assert (np.isclose(oblate.area(), 78.04694433010546, 0., 1e-15))
    assert (np.isclose(prolate.area(), 48.326479487738396, 0., 1e-15))
    assert (ds['is_intersection'].values == prolate.is_intersection(r2))


def test_disk():
    # 1) particular cases
    disk = gc.Disk(radius=1.5)
    ds = gc.calc_intersection(disk, gc.Ray(gc.Point(5.,0.,0.), gc.Vector(-1.,0.,0.)))
    assert (not ds['is_intersection'])
    ds = gc.calc_intersection(disk, gc.Ray(gc.Point(0.,0.,10.), gc.Vector(0.,0.,-1.), maxt=5.))
    assert (not ds['is_intersection'])

    disk = gc.Disk(radius=1., phi_max=90.)
    ds = gc.calc_intersection(disk, gc.Ray(gc.Point(0.8,-1e-5,10.), gc.Vector(0.,0.,-1.)))
    assert (not ds['is_intersection'])
    ds = gc.calc_intersection(disk, gc.Ray(gc.Point(0.8,1e-5,10.), gc.Vector(0.,0.,-1.)))
    assert (ds['is_intersection'])
    ds = gc.calc_intersection(disk, gc.Ray(gc.Point(-1e-5,0.8,10.), gc.Vector(0.,0.,-1.)))
    assert (not ds['is_intersection'])
    ds = gc.calc_intersection(disk, gc.Ray(gc.Point(+1e-5,0.8,10.), gc.Vector(0.,0.,-1.)))
    assert (ds['is_intersection'])

    # 2) general cases
    roty_90 = gc.get_rotateY_tf(90.)
    disk = gc.Disk(radius=1.5, z_height=5., oTw=roty_90)
    annulus = gc.Disk(radius=1.5, inner_radius=0.8, z_height=5., oTw=roty_90)
    d = gc.Vector(-1.,0.,0.)
    r1 = gc.Ray(gc.Point(10.,0.,1.2), d)
    r2 = gc.Ray(gc.Point(10.,0.,0.2), d)
    r3 = gc.Ray(gc.Point(10.,0.,1.6), d)

    # 2a) disk tests
    ds_disk = gc.calc_intersection(disk, r1)
    p = ds_disk['phit'].values
    assert (np.all(np.isclose(p, r1(disk.z_height).to_numpy(), 0., 1e-15)))
    assert (ds_disk['is_intersection'].values == disk.is_intersection(r1))
    ds_disk = gc.calc_intersection(disk, r2)
    p = ds_disk['phit'].values
    assert (np.all(np.isclose(p, r2(disk.z_height).to_numpy(), 0., 1e-15)))
    assert (ds_disk['is_intersection'].values == disk.is_intersection(r2))
    ds_disk = gc.calc_intersection(disk, r3)
    assert (not ds_disk['is_intersection'])
    assert (ds_disk['is_intersection'].values == disk.is_intersection(r3))

    # 2b) annulus tests
    ds_annulus = gc.calc_intersection(annulus, r1)
    p = ds_annulus['phit'].values
    assert (np.all(np.isclose(p, r1(annulus.z_height).to_numpy(), 0., 1e-15)))
    assert (ds_annulus['is_intersection'].values == annulus.is_intersection(r1))
    ds_annulus = gc.calc_intersection(annulus, r2)
    assert (not ds_annulus['is_intersection'])
    assert (ds_annulus['is_intersection'].values == annulus.is_intersection(r2))
    ds_annulus = gc.calc_intersection(annulus, r3)
    assert (not ds_annulus['is_intersection'])
    assert (ds_annulus['is_intersection'].values == annulus.is_intersection(r3))


@pytest.mark.parametrize('quadric', Q1)
def test_quadric_1d_arr(quadric):
    x_, y_, z_ = np.meshgrid(np.linspace(-1.5, 1.5, 10, np.float64),
                         np.linspace(-1.5, 1.5, 10, np.float64),
                         5., indexing='ij')
    o_set_arr = np.vstack((x_.ravel(), y_.ravel(), z_.ravel())).T
    nrays = o_set_arr.shape[0]
    d_set_arr = np.zeros_like(o_set_arr)
    d_set_arr[:,0] = 0.
    d_set_arr[:,1] = 0.
    d_set_arr[:,2] = -1.
    o_set = gc.Point(o_set_arr)
    d_set = gc.Vector(d_set_arr)
    r_set = gc.Ray(o_set, d_set)

    ds = quadric.intersect(r_set, ds_output=True)

    is_int_1d = np.full((nrays), True, dtype=bool)
    t_1d = np.zeros((nrays), dtype=np.float64)
    p_1d = np.zeros((nrays,3), dtype=np.float64)
    n_1d = np.zeros_like(p_1d)
    dpdu_1d = np.zeros_like(p_1d)
    dpdv_1d = np.zeros_like(p_1d)
    u_1d = np.zeros_like(t_1d)
    v_1d = np.zeros_like(t_1d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for ir in range (0, nrays):
        res_sca = quadric.intersect(list_rays[ir], ds_output=False)
        if (res_sca[2] is not None):
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[ir] = ds_sca['is_intersection'].values
            t_1d[ir] = ds_sca['thit'].values
            p_1d[ir,:] = ds_sca['phit'].values
            n_1d[ir,:] = ds_sca['nhit'].values
            dpdu_1d[ir,:] = ds_sca['dpdu'].values
            dpdv_1d[ir,:] = ds_sca['dpdv'].values
            u_1d[ir] = ds_sca['u'].values
            v_1d[ir] = ds_sca['v'].values
        else:
            is_int_1d[ir] = False
            t_1d[ir] = None
            p_1d[ir,:] = None
            n_1d[ir,:] = None
            dpdu_1d[ir,:] = None
            dpdv_1d[ir,:] = None
            u_1d[ir] = None
            v_1d[ir] = None
    
    assert (np.array_equal(ds['is_intersection'].values, is_int_1d, equal_nan=True))
    assert (np.array_equal(ds['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds['phit'].values, p_1d, equal_nan=True))
    assert (np.allclose(ds['nhit'].values, n_1d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds['u'].values, u_1d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds['v'].values, v_1d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds['dpdu'].values, dpdu_1d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds['dpdv'].values, dpdv_1d, 0., 1e-14, equal_nan=True))

    thit, is_int = quadric.is_intersection_t(r_set)

    assert (np.array_equal(ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(ds['thit'].values, thit, equal_nan=True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Vector, Point, Normal
import math
import numpy as np


def dot(a, b):
    """
    The dot/scalar product

    Definition of the dot product:

    - :math:`(a . b) = a.x*b.x + a.y*b.y + a.z*b.z`
    - :math:`(a . b) = ||a|| * ||b|| * cos(θ)`

    Parameters
    ----------
    a : Vector | Normal
        The first vector(s) or normal(s) used for the dot product
    b : Vector | Normal
        The second vector(s) or normal(s) used for the dot product
    
    Returns
    -------
    out : float | 1-D ndarray
        The result(s) of the dot product i.e. sum of products

    Examples
    --------
    >>> import geoclide as gc
    >>> a = gc.Vector(0., 0., 1.)
    >>> b = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    >>> gc.dot(a,b)
    0.7071067811865476
    """
    if (isinstance(a, Vector) or isinstance(a, Normal)) and \
       (isinstance(b, Vector) or isinstance(b, Normal)):
        return (a.x*b.x + a.y*b.y + a.z*b.z)
    else:
        raise ValueError('Only Vector or Normal parameters are accepted')


def cross(a, b):
    """
    The cross product

    - .. warning:: The cross product of 2 normals is not allowed

    Definition of the cross product:

    - :math:`(a × b) = ((a.y*b.z)-(a.z*b.y))x̂ + ((a.z*b.x)-(a.x*b.z))ŷ + ((a.x*b.y)-(a.y*b.x))`
    - :math:`(a × b) = ||a||*||b||*sin(θ)`

        where x̂, ŷ and ẑ are the unitary vectors respectively in axes x, y and z

    Parameters
    ----------
    a : Vector | Normal
        The first vector(s) or normal(s) used for the cross product
    b : Vector | Normal
        The second vector(s) or normal(s) used for the cross product
    
    Returns
    -------
    out : Vector
        The result(s) of the cross product

    Examples
    --------
    >>> import geoclide as gc
    >>> a = gc.Vector(0.,0.,1.)
    >>> b = gc.Vector(1.,0.,0.)
    >>> gc.cross(a,b)
    Vector(0.0, 1.0, 0.0)
    """
    if ( (isinstance(a, Vector) and isinstance(b, Vector)) or 
         (isinstance(a, Vector) and isinstance(b, Normal)) or 
         (isinstance(a, Normal) and isinstance(b, Vector)) ):
        return Vector((a.y*b.z)-(a.z*b.y), (a.z*b.x)-(a.x*b.z), (a.x*b.y)-(a.y*b.x))
    elif isinstance(a, Normal) and isinstance(b, Normal):
        raise ValueError('Only 1 Normal is tolerated not 2')
    else:
        raise ValueError('Only Vector or Normal parameters are accepted')
    

def normalize(v):
    """
    Normalize a vector/normal or a set of vectors/normals

    Parameters
    ----------
    v : Vector | Normal
        The vector(s) or normal(s) to be normalized
    
    Returns
    -------
    out : Vector | Normal
        The normalized vector(s)/normal(s)

    Examples
    --------
    >>> import geoclide as gc
    >>> v = gc.Vector(1.,0.,1.)
    >>> gc.normalize(v)
    Vector(0.7071067811865475, 0.0, 0.7071067811865475)
    """
    if isinstance(v, Vector) or isinstance(v, Normal):
        return v / v.length()
    else:
        raise ValueError('The parameter v must be a Vector or Normal')


def coordinate_system(v1, method="m2"):
    """
    Create orthogonal coordinate system(s) from a vector/set of vectors

    Parameters
    ----------
    v1 : Vector
        The base vector(s) used to create the orthogonal coordinate system(s)
    method: str, optional
        Default is 'm2' (method from pbrt v4), other choice is 'm1' (pbrt v2 and v3)
    
    Returns
    -------
    v2 : Vector
        The second vector(s) of the orthogonal coordinate system(s)
    v3 : Vector
        The third vector(s) of the orthogonal coordinate system(s)
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(0., 0., 1.)
    >>> v2, v3 = gc.coordinate_system(v1)
    (Vector(1.0, -0.0, -0.0), Vector(-0.0, 1.0, -0.0))
    """
    if not isinstance(v1, Vector):
        raise ValueError("The parameter v1 must be a Vector")

    # used in pbrt v4    
    if (method == "m2"):
        if (isinstance(v1.x, np.ndarray)):
            sign = np.ones_like(v1.x)
            sign[v1.z<=0] = -1
        else:
            if v1.z > 0:sign = 1.
            else: sign =-1.
        term_1 = -1. / (sign + v1.z)
        term_2 = v1.x * v1.y * term_1
        v2 = Vector(1 + sign * (v1.x*v1.x) * term_1, sign * term_2, -sign * v1.x)
        v3 = Vector(term_2, sign + (v1.y*v1.y) * term_1, -v1.y)
    # used in pbrt v2 and v3
    elif (method == "m1"):
        if (isinstance(v1.x, np.ndarray)):
            v2 = np.zeros((len(v1.x), 3), np.float64)
            cond = np.abs(v1.x) > np.abs(v1.y)
            not_cond = np.logical_not(cond)
            any_cond = np.any(cond)
            any_not_cond = np.any(not_cond)
            if any_cond:
                v2[cond,0] =  (1./np.sqrt(v1.x*v1.x + v1.z*v1.z)) * -v1.z
                v2[cond,2] =  (1./np.sqrt(v1.x*v1.x + v1.z*v1.z)) * v1.x
            if any_not_cond:
                v2[not_cond,1] = (1./np.sqrt(v1.x*v1.x + v1.z*v1.z)) * v1.z
                v2[not_cond,2] = (1./np.sqrt(v1.x*v1.x + v1.z*v1.z)) * -v1.y
            v2 = Vector(v2)
        else:
            if (abs(v1.x) > abs(v1.y)):
                invLen = 1/ math.sqrt(v1.x*v1.x + v1.z*v1.z)
                v2 = Vector(-v1.z*invLen, 0, v1.x*invLen)
            else:
                invLen = 1/ math.sqrt(v1.y*v1.y + v1.z*v1.z)
                v2 = Vector(0, v1.z*invLen, -v1.y*invLen)
        v3 = cross(v1, v2)
    else:
        raise ValueError("Only 2 choices for parameter method: 'm1' or 'm2'")
    
    return v2, v3


def distance(p1, p2):
    """
    Compute the distance(s) between 2 points/set of points

    Parameters
    ----------
    p1 : Point
        The first point(s)
    p2 : Point
        The second point(s)

    Returns
    -------
    out : float | 1-D ndarray
        The distance(s) between the 2 points/set of points
    
    Examples
    --------
    >>> p1 = gc.Point(0., 0., 0.)
    >>> p2 = gc.Point(0., 0., 10.)
    >>> gc.distance(p1,p2)
    10.0
    >>> p1 = gc.Point(1., 2., 1.9)
    >>> p2 = gc.Point(5., 15., 3.)
    >>> gc.distance(p1,p2)
    13.64587849865299
    """
    if isinstance(p1, Point) and isinstance(p2, Point):
        return (p1 - p2).length()
    else:
        raise ValueError('Only Point parameters are accepted')


def face_forward(a, b):
    """
    Flip the vector(s)/normal(s) a if the vector(s)/normal(s) b is/are in the opposite 
    direction(s)

    It can be useful to flip a surface normal so that it lies in the same
    hemisphere as a given vector.

    Parameters
    ----------
    a : Vector | Normal
        The vector(s) or normal(s) to potentially flip
    b : Vector | Normal
        The base vector(s) or normal(s) used for the flip

    Returns
    -------
    out : Vector | Normal
        The potentially flipped vector(s) or normal(s)

    Examples
    --------
    >>> import geoclide as gc
    >>> n1 = gc.Normal(1., 0., 0.)
    >>> v1 = gc.Vector(-1., 0., 0.)
    >>> gc.face_forward(v1, n1)
    Vector(1.0, -0.0, -0.0)
    """
    if (isinstance(a, Vector) or isinstance(a, Normal)) and \
    (isinstance(b, Vector) or isinstance(b, Normal)):
        if isinstance(a.x, np.ndarray):
            a_bis = a.to_numpy()
            cond = dot(a, b) < 0
            if np.any(cond): a_bis[cond,:] *= -1
            if(isinstance(a, Vector)): return Vector(a_bis)
            else: return Normal(a_bis)
        elif isinstance(b.x, np.ndarray):
            a_bis = np.zeros_like(b.to_numpy())
            a_bis[:,0] = a.x
            a_bis[:,1] = a.y
            a_bis[:,2] = a.z
            cond = dot(a, b) < 0
            if np.any(cond): a_bis[cond,:] *= -1
            if(isinstance(a, Vector)): return Vector(a_bis)
            else: return Normal(a_bis)
        else:
            return (a*-1) if (dot(a, b) < 0) else a
    else:
        raise ValueError('Only Vector or Normal parameters are accepted')


def vmax(a):
    """
    Get the largest components value(s) of the vector(s)/point(s)/normal(s)

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector(s)/point(s)/normal(s) used
    
    Returns
    -------
    out: float | 1-D ndarray
        The largest vector(s)/point(s)/normal(s) value(s)
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.max(v1)
    3
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        if (isinstance(a.x, np.ndarray)):
            return np.maximum(a.x, np.maximum(a.y, a.z))
        else:
            return max(a.x, max(a.y, a.z))

    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')


def vmin(a):
    """
    Get the smallest components value(s) of the vector(s)/point(s)/normal(s)

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector(s)/point(s)/normal(s) used
    
    Returns
    -------
    out: float | 1-D ndarray
        The smallest vector(s)/point(s)/normal(s) value(s)
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.min(v1)
    1
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        if (isinstance(a.x, np.ndarray)):
            return np.minimum(a.x, np.minimum(a.y, a.z))
        else:
            return min(a.x, min(a.y, a.z))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')
    

def vargmax(a):
    """
    Get the index/indices of the vector(s)/point(s)/normal(s) components with 
    the largest value(s)

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector(s)/point(s)/normal(s) used
    
    Returns
    -------
    out: int | 1-D ndarray
        The index/indices of the largest vector(s)/point(s)/normal(s) value(s)
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.argmax(v1)
    1
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        if (isinstance(a.x, np.ndarray)):
            a_bis = np.zeros_like(a.x, dtype=np.int32)
            c1 = a.x>a.y
            not_c1 = np.logical_not(c1)
            c2 = a.x<=a.z
            c3 = a.y>a.z
            not_c3 = np.logical_not(c3)
            a_bis[np.logical_and(c1, c2)] = 2
            a_bis[np.logical_and(not_c1, c3)] = 1
            a_bis[np.logical_and(not_c1, not_c3)] = 2
            return a_bis
        else:
            return (0 if a.x>a.z else 2) if (a.x>a.y) else (1 if a.y>a.z else 2)
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')


def vargmin(a):
    """
    Get the index/indices of the vector(s)/point(s)/normal(s) components with 
    the smallest value(s)

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector(s)/point(s)/normal(s) used
    
    Returns
    -------
    out: int | 1-D ndarray
        The index/indices of the smallest vector(s)/point(s)/normal(s) value(s)
    
    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2.,3.,1.)
    >>> gc.argmin(v1)
    2
    """
    if isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal):
        if (isinstance(a.x, np.ndarray)):
            a_bis = np.zeros_like(a.x, dtype=np.int32)
            c1 = a.x<a.y
            not_c1 = np.logical_not(c1)
            c2 = a.x>=a.z
            c3 = a.y<a.z
            not_c3 = np.logical_not(c3)
            a_bis[np.logical_and(c1, c2)] = 2
            a_bis[np.logical_and(not_c1, c3)] = 1
            a_bis[np.logical_and(not_c1, not_c3)] = 2
            return a_bis
        else:
            return (0 if a.x<a.z else 2) if (a.x<a.y) else (1 if a.y<a.z else 2)
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')
    

def vabs(a):
    """
    Calculate the absolute value(s) of each component of the vector(s)/point(s)/normal(s)

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector(s)/point(s)/normal(s) used
    
    Returns
    -------
    out: Vector | Point | Normal | 1-D ndarray
        The vector(s)/point(s)/normal(s) with absolute values
    """
    if isinstance(a, Vector):
        if (isinstance(a.x, np.ndarray)):
            return Vector(np.abs(a.x), np.abs(a.y), np.abs(a.z))
        else:
            return Vector(abs(a.x), abs(a.y), abs(a.z))
    elif isinstance(a, Point):
        if (isinstance(a.x, np.ndarray)):
            return Point(np.abs(a.x), np.abs(a.y), np.abs(a.z))
        else:
            return Point(abs(a.x), abs(a.y), abs(a.z))
    elif isinstance(a, Normal):
        if (isinstance(a.x, np.ndarray)):
            return Normal(np.abs(a.x), np.abs(a.y), np.abs(a.z))
        else:
            return Normal(abs(a.x), abs(a.y), abs(a.z))
    else:
        raise ValueError('Only a Vector, a Point or a Normal parameter is accepted')


def permute(a, ix=None, iy=None, iz=None):
    """
    Permutes the vector(s)/point(s)/normal(s) x, y, z values according to the given indices

    Parameters
    ----------
    a : Vector | Point | Normal
        The vector(s), point(s) or normal(s) used for permutation
    ix : int | np.ndarray | list, optional
        The index/indices of the value(s) we want to keep as a remplacement for the x component(s)
    iy : int | np.ndarray | list, optional
        The index/indices of the value(s) we want to keep as a remplacement for the y component(s)
    iz : int | np.ndarray | list, optional
        The index/indices of the value(s) we want to keep as a remplacement for the z component(s)
    
    Returns
    -------
    out : Vector | Point | Normal
        The vector(s)/point(s)/normal(s) after the permute operation

    Examples
    --------
    >>> import geoclide as gc
    >>> v1 = gc.Vector(2., 3., 1.)
    >>> gc.permute(v1, 1, 0, 2)
    Vector(3.0, 2.0, 1.0)
    >>> gc.permute(v1, 1, 0, 2])
    Vector(3.0, 2.0, 1.0)
    >>> p_set = np.array([[0.,0.,3.], [1.,0.,0.], [-0.5,0.,0.], [0.,-3.,0.]])
    >>> p_set = gc.Point(p_set)
    >>> gc.permute(p_set, [2,0,0,1], 1, 2).to_numpy()
    array([[ 3. ,  0. ,  3. ],
           [ 1. ,  0. ,  0. ],
           [-0.5,  0. ,  0. ],
           [ 0. , -3. ,  0. ]])
    """
    if not(isinstance(a, Vector) or isinstance(a, Point) or isinstance(a, Normal)):
        raise NameError('The parameter a must be a Vector or Point or Normal')
    if ix is None : ix = int(0)
    if iy is None : iy = int(1)
    if iz is None : iz = int(2)

    if ( isinstance(ix, np.ndarray) or isinstance(ix, list) or
         isinstance(iy, np.ndarray) or isinstance(iy, list) or
         isinstance(iz, np.ndarray) or isinstance(iy, list)):
        a_arr = a.to_numpy()
    if ( isinstance(ix, np.ndarray) or isinstance(ix, list) ): ax = a_arr[np.arange(len(ix)), ix]
    else: ax=a[ix]
    if ( isinstance(iy, np.ndarray) or isinstance(iy, list) ): ay = a_arr[np.arange(len(iy)), iy]
    else: ay=a[iy]
    if ( isinstance(iz, np.ndarray) or isinstance(iz, list) ): az = a_arr[np.arange(len(iz)), iz]
    else: az=a[iz]
    return a.__class__(ax, ay, az)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Vector
from geoclide.vecope import normalize
from geoclide.transform import get_rotateY_tf, get_rotateZ_tf
from geoclide.mathope import clamp
import numpy as np
import math
from warnings import warn

# TODO add ds_output option
def ang2vec(theta, phi, vec_view='zenith', diag_calc=False):
    """
    Convert a direction/directions described by 2 angles/set of 2 angles into a 
    direction/directions described by a vector/vectors

    - direct orthogonal coordinate system where z is pointing upwards

    Parameters
    ----------
    theta : float | 1D array
        The polar angle(s) in degrees, starting at z+ in the zx plane and going 
        in the trigonometric direction around the y axis

    phi : float | 1D array
        The azimuthal angle(s) in degrees, starting at x+ in the xy plane and going in 
        the trigonométric direction around the z axis

    vec_view : str, optional
        Two choices (concerning intial direction at theta=phi=0): 'zenith' (i.e. pointing above) or 
        'bellow' (i.e. pointing bellow)
    diag_calc : bool, optional
            Perform diagonal calculations, v(i) is calculated using theta(i) and phi(i)

    Returns
    -------
    v : Vector
        The direction(s) described by a vector

    Notes
    -----
    In case both theta and phi are 1-D ndarrays, the vectors are ordered in phi-major order

    Examples
    --------
    >>> import geoclide as gc
    >>> th = 30.
    >>> ph = 0.
    >>> v1 = gc.ang2vec(theta=th, phi=ph, vec_view='zenith')
    >>> v1
    Vector(0.49999999999999994, 0.0, 0.8660254037844387)
    >>> v2 = gc.ang2vec(theta=th, phi=ph, vec_view='nadir')
    >>> v2
    Vector(-0.49999999999999994, 0.0, -0.8660254037844387)
    """
    if (vec_view == "zenith"): # initial vector is facing zenith (pointing above)
        v = Vector(0., 0., 1.)
    elif (vec_view == "nadir"): # initial vector is facing nadir (pointing bellow)
        v = Vector(0., 0., -1.)
    else:
        raise ValueError("The value of vec_view parameter must be: 'zenith' or 'nadir")
    
    if isinstance(theta, np.ndarray) or isinstance(phi, np.ndarray): flatten = True
    else : flatten = False

    v = get_rotateY_tf(theta)(v, flatten=flatten)
    v = get_rotateZ_tf(phi)(v, flatten=flatten, diag_calc=diag_calc)
    v = normalize(v)
    
    return v


def vec2ang(v, vec_view='zenith', acc=1e-6):
    """
    Convert a direction/directions described by a vector/vectors into 
    a direction/directions described by 2 angles/set of 2 angles

    - direct orthogonal coordinate system where z is pointing upwards

    Parameters
    ----------
    v : Vector
        The direction(s) described by a vector
    vec_view : str, optional
        Two choices (concerning intial direction at theta=phi=0): 'zenith' (i.e. pointing above) or 
        'nadir' (i.e. pointing bellow)
    acc : float, optional
        The tolerance for numerical errors. Default is 1e-6.

    Returns
    -------
    theta : float | 1-D ndarray
        The polar angle(s) in degrees, starting at z+ in the zx plane and going 
        in the trigonometric direction around the y axis
    phi : float | 1-D ndarray
        The azimuthal angle(s) in degrees, starting at x+ in the xy plane and going in 
        the trigonométric direction around the z axis

    Examples
    --------
    >>> import geoclide as gc
    >>> th = 30.
    >>> ph = 0.
    >>> v1 = gc.ang2vec(theta=th, phi=ph, vec_view='zenith')
    >>> v1
    Vector(0.49999999999999994, 0.0, 0.8660254037844387)
    >>> theta, phi = gc.vec2ang(v1, vec_view='zenith')
    >>> theta, phi
    (29.999999999999993, 0.0)
    >>> v2 = gc.ang2vec(theta=th, phi=ph, vec_view='nadir')
    >>> v2
    Vector(-0.49999999999999994, 0.0, -0.8660254037844387)
    >>> theta, phi = gc.vec2ang(v1, vec_view='nadir')
    >>> theta, phi
    (29.999999999999993, 0.0)
    """
    if (not isinstance(v, Vector)):
        raise ValueError('The parameter v must be a vector')
    
    if (vec_view == "zenith"): # initial vector is facing zenith (pointing above)
        pass
    elif (vec_view == "nadir"): # initial vector is facing nadir (pointing bellow)
        v = -v # by doing that, we can keep a v_ini facing upwards
    else:
        raise ValueError("The value of vec_view parameter must be: 'zenith' or 'nadir")
    
    v = normalize(v) # ensure v is normalized

    if isinstance(v.x, np.ndarray):
        nv = len(v.x)
        v_arr = v.to_numpy()
        v_ini_arr = np.zeros_like(v_arr)
        v_ini_arr[:,2] = 1.
        v_ini = Vector(v_ini_arr)
        theta = np.full(nv, np.nan, dtype=np.float64)
        phi = theta.copy()
        
        c1 = np.all(np.isclose(v_arr, v_ini_arr, 0., acc), axis=1)
        not_resolved = np.logical_not(c1)
        if not_resolved.sum() == 0 : return theta, phi

        for icase in range (1, 6):
            if icase == 1:
                roty_rad = np.arccos(v.z)
                c_case_bis = np.logical_not(np.logical_and(v.x == 0, roty_rad == 0))
                cosphi = np.zeros(nv, dtype=np.float64)
                cosphi[c_case_bis] =  np.clip(v.x[c_case_bis]/np.sin(roty_rad[c_case_bis]), -1., 1.)
                rotz_rad = np.arccos(cosphi)
                theta_bis = np.degrees(roty_rad)
                phi_bis = np.degrees(rotz_rad)
            elif icase == 2:
                rotz_rad = -np.arccos(cosphi)
                phi_bis = np.degrees(rotz_rad)
            elif icase == 3:
                roty_rad = -np.arccos(v.z)
                cosphi = np.zeros(nv, dtype=np.float64)
                cosphi[c_case_bis] =  np.clip(v.x[c_case_bis]/np.sin(roty_rad[c_case_bis]), -1., 1.)
                rotz_rad = np.arccos(cosphi)
                theta_bis = np.degrees(roty_rad)
                phi_bis = np.degrees(rotz_rad)
            elif icase == 4:
                rotz_rad = -np.arccos(cosphi)
                phi_bis = np.degrees(rotz_rad)
            else:
                warn('No rotation has been found for some (or all) vectors!', Warning)
                return theta, phi

            rotzy = get_rotateZ_tf(phi_bis)*get_rotateY_tf(theta_bis)
            v_ini_rotated = normalize(rotzy(v_ini, flatten=True, diag_calc=True))
            c_tmp = np.all(np.isclose(v_arr, v_ini_rotated.to_numpy(), 0., acc), axis=1)
            c_tmp_bis = np.logical_and(not_resolved, c_tmp)
            theta[c_tmp_bis] = theta_bis[c_tmp_bis]
            phi[c_tmp_bis] = phi_bis[c_tmp_bis]
            not_resolved = np.logical_and(not_resolved, np.logical_not(c_tmp))
            if not_resolved.sum() == 0 : return theta, phi
    else: # if only 1 vector
        v_ini = Vector(0., 0., 1.)

        # In case v = v_ini -> no rotations
        if (np.all(np.isclose(v.to_numpy()-v_ini.to_numpy(), 0., 0., acc))):
            return 0., 0.
        
        for icase in range (1, 6):
            if icase == 1:
                roty_rad = math.acos(v.z)
                if (v.x == 0 and roty_rad == 0): cosphi = 0.
                else: cosphi = clamp(v.x/math.sin(roty_rad), -1., 1.)
                rotz_rad = math.acos(cosphi)
            elif(icase == 2):
                roty_rad = math.acos(v.z)
                if (v.x == 0 and roty_rad == 0): cosphi = 0.
                else: cosphi = clamp(v.x/math.sin(roty_rad), -1., 1.)
                rotz_rad = -math.acos(cosphi)
            elif(icase == 3):
                roty_rad = -math.acos(v.z)
                if (v.x == 0 and roty_rad == 0): cosphi = 0.
                else: cosphi = clamp(v.x/math.sin(roty_rad), -1., 1.)
                rotz_rad = math.acos(cosphi)
            elif(icase == 4):
                roty_rad = -math.acos(v.z)
                if (v.x == 0 and roty_rad == 0): cosphi = 0.
                else: cosphi = clamp(v.x/math.sin(roty_rad), -1., 1.)
                rotz_rad = -math.acos(cosphi)
            else:
                raise NameError('No rotation has been found!')
            
            theta = math.degrees(roty_rad)
            phi = math.degrees(rotz_rad)
            rotzy = get_rotateZ_tf(phi)*get_rotateY_tf(theta)
            v_ini_rotated = normalize(rotzy(v_ini))

            if (np.all(np.isclose(v.to_numpy()-v_ini_rotated.to_numpy(), 0., 0., acc))):
                break
        
        return theta, phi



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Ray, BBox
from geoclide.quadrics import Sphere, Spheroid, Disk
from geoclide.trianglemesh import Triangle, TriangleMesh
from geoclide.shapes import Shape
import xarray as xr


def calc_intersection(shape, r, **kwargs):
    """
    Performs intersection test between a shape and a ray and returns dataset

    Parameters
    ----------
    shape : BBox | Sphere | Spheroid | Disk | Triangle | TriangleMesh
        The shape used for the intersection(s)
    r : Ray
        The ray(s) used for the intersection(s)
    **kwargs
        The keyword arguments are passed on to intersect method. The ds_output parameter 
        is forced here to always be True.

    Returns
    -------
    out : xr.Dataset
        Look-up table with the intersection information
    
    Examples
    --------
    >>> import geoclide as gc
    >>> sphere = gc.Sphere(radius=1.) # sphere of radius 1
    >>> bbox = gc.BBox(p1=gc.Point(0., 0., 0.), p2=gc.Point(1.,1.,1.))
    >>> ray = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
    >>> ds_sphere = gc.calc_intersection(sphere, ray)
    >>> ds_sphere
    <xarray.Dataset> Size: 753B
    Dimensions:          (xyz: 3, dim_0: 4, dim_1: 4)
    Coordinates:
    * xyz              (xyz) int64 24B 0 1 2
    Dimensions without coordinates: dim_0, dim_1
    Data variables: (12/20)
        o                (xyz) float64 24B -2.0 0.0 0.8
        d                (xyz) float64 24B 1.0 0.0 0.0
        mint             int64 8B 0
        maxt             float64 8B inf
        is_intersection  bool 1B True
        thit             float64 8B 1.4
        ...               ...
        z_max            float64 8B 1.0
        phi_max          float64 8B 360.0
        wTo_m            (dim_0, dim_1) float64 128B 1.0 0.0 0.0 0.0 ... 0.0 0.0 1.0
        wTo_mInv         (dim_0, dim_1) float64 128B 1.0 0.0 0.0 0.0 ... 0.0 0.0 1.0
        oTw_m            (dim_0, dim_1) float64 128B 1.0 0.0 0.0 0.0 ... 0.0 0.0 1.0
        oTw_mInv         (dim_0, dim_1) float64 128B 1.0 0.0 0.0 0.0 ... 0.0 0.0 1.0
    >>> ds_box = gc.calc_intersection(bbox, ray)
    >>> ds_bbox
    <xarray.Dataset> Size: 169B
    Dimensions:          (xyz: 3)
    Coordinates:
    * xyz              (xyz) int64 24B 0 1 2
    Data variables:
        is_intersection  bool 1B True
        o                (xyz) float64 24B -2.0 0.0 0.8
        d                (xyz) float64 24B 1.0 0.0 0.0
        mint             int64 8B 0
        maxt             float64 8B inf
        pmin             (xyz) float64 24B 0.0 0.0 0.0
        pmax             (xyz) float64 24B 1.0 1.0 1.0
        thit             float64 8B 2.0
        phit             (xyz) float64 24B 0.0 0.0 0.8
    """
    if (not isinstance(r, Ray)):
        raise ValueError('The parameter r1 must a Ray')

    if 'ds_output' in kwargs: kwargs.pop('ds_output', False)
    if (isinstance(shape, BBox)) or issubclass(shape.__class__, Shape):
        ds = shape.intersect(r, ds_output=True, **kwargs)
    else:
        raise ValueError('The only supported shape are: BBox, Sphere, Spheroid, Disk, ' +
                         'Triangle and TriangleMesh')

    if (isinstance(shape, BBox)):
        ds['pmin'] = xr.DataArray(shape.pmin.to_numpy(), dims='xyz')
        ds['pmin'].attrs = {'type': 'Point', 'description':'the x, y and z components of the pmin BBox attribut'}
        ds['pmax'] = xr.DataArray(shape.pmax.to_numpy(), dims='xyz')
        ds['pmax'].attrs = {'type': 'Point', 'description':'the x, y and z components of the pmax BBox attribut'}
    if (isinstance(shape, Sphere)):
        ds['radius'] = shape.radius
        ds['radius'].attrs = {'description':'the sphere radius attribut'}
        ds['z_min'] = shape.zmin
        ds['z_min'].attrs = {'description':'the sphere zmin attribut'}
        ds['z_max'] = shape.zmax
        ds['z_max'].attrs = {'description':'the sphere zmax attribut'}
        ds['phi_max'] = shape.phi_max
        ds['phi_max'].attrs = {'unit':'Degree', 'description':'the sphere phi_max attribut'}
    if (isinstance(shape, Spheroid)):
        ds['radius_xy'] = shape.alpha
        ds['radius_xy'].attrs = {'description':'the equatorial radius of the spheroid (alpha attribut)'}
        ds['radius_z'] = shape.gamma
        ds['radius_z'].attrs = {'description':'the distance between the spheroid center and pole (gamma attribut)'}
    if (isinstance(shape, Disk)):
        ds['radius'] = shape.radius
        ds['radius'].attrs = {'description':'the radius of the disk'}
        ds['inner_radius'] = shape.inner_radius
        ds['inner_radius'].attrs = {'description':'the inner radius of the disk (if > 0 -> annulus case)'}
        ds['phi_max'] = shape.phi_max
        ds['phi_max'].attrs = {'unit':'Degree', 'description':'the disk phi_max attribut'}
        ds['z_height'] = shape.z_height
        ds['z_height'].attrs = {'description':'the disk z_height attribut'}
    if (isinstance(shape, Triangle)):
        ds['p0'] = xr.DataArray(shape.p0.to_numpy(), dims='xyz')
        ds['p0'].attrs = {'description': 'the triangle p0 attribut'}
        ds['p1'] = xr.DataArray(shape.p1.to_numpy(), dims='xyz')
        ds['p1'].attrs = {'description': 'the triangle p1 attribut'}
        ds['p2'] = xr.DataArray(shape.p2.to_numpy(), dims='xyz')
        ds['p2'].attrs = {'description': 'the triangle p2 attribut'}
    if (isinstance(shape, TriangleMesh)):
        ds['vertices'] = xr.DataArray(shape.vertices, dims=['nvertices', 'xyz'])
        ds['vertices'].attrs = {'description': 'The vertices xyz coordinates.'}
        ds['faces'] = xr.DataArray(shape.faces, dims=['ntriangles', 'p0p1p2'])
        ds['faces'].attrs = {'description': 'For each triangle, the index of vertices point p0, p1 and p2 (from variable v).'}
        ds.attrs.update({'ntriangles': shape.ntriangles, 'nvertices' : shape.nvertices})
    if (not isinstance(shape, BBox)):
        ds['wTo_m'] = xr.DataArray(shape.wTo.m)
        ds['wTo_m'].attrs = {'description':'the transformation matrix of the ' + str(ds.attrs['shape']).lower() + ' wTo attribut'}
        ds['wTo_mInv'] = xr.DataArray(shape.wTo.mInv)
        ds['wTo_mInv'].attrs = {'description':'the inverse transformation matrix of the ' + str(ds.attrs['shape']).lower() + ' wTo attribut'}
        ds['oTw_m'] = xr.DataArray(shape.oTw.m)
        ds['oTw_m'].attrs = {'description':'the transformation matrix of the ' + str(ds.attrs['shape']).lower() + ' oTw attribut'}
        ds['oTw_mInv'] = xr.DataArray(shape.oTw.mInv)
        ds['oTw_mInv'].attrs = {'description':'the inverse transformation matrix of the ' + str(ds.attrs['shape']).lower() +' oTw attribut'}

    return ds


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.mathope import gamma_f32, gamma_f64
import math

VERSION = '3.0.3'

GAMMA2_F32 = gamma_f32(2)
GAMMA3_F32 = gamma_f32(3)
GAMMA5_F32 = gamma_f32(5)

GAMMA2_F64 = gamma_f64(2)
GAMMA3_F64 = gamma_f64(3)
GAMMA5_F64 = gamma_f64(5)

TWO_PI = math.pi*2.


from geoclide.basic import Vector, Point, Normal, Ray, BBox, get_common_vertices, get_common_face
from geoclide.vecope import dot, cross, normalize, coordinate_system, distance, face_forward, \
    vmax, vmin, vargmax, vargmin, vabs, permute
from geoclide.mathope import clamp, quadratic, gamma_f32, gamma_f64
from geoclide.transform import Transform, get_translate_tf, get_scale_tf, \
    get_rotateX_tf, get_rotateY_tf, get_rotateZ_tf, get_rotate_tf, get_inverse_tf
from geoclide.quadrics import Sphere, Spheroid, Disk
from geoclide.intersection import calc_intersection
from geoclide.trianglemesh import Triangle, TriangleMesh, create_sphere_trianglemesh, \
    read_trianglemesh, create_disk_trianglemesh
from geoclide.constante import VERSION, GAMMA2_F32, GAMMA2_F64, GAMMA3_F32, GAMMA3_F64, GAMMA5_F32, GAMMA5_F64, TWO_PI
from geoclide.advancedvecope import ang2vec, vec2ang



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Vector, Point, Normal, Ray
from geoclide.vecope import normalize, cross, face_forward
import numpy as np
from geoclide.transform import Transform
import xarray as xr
from datetime import datetime
from geoclide.constante import VERSION


class Shape(object):
    '''
    :meta private:
    
    The parent class of the different shape classes
    '''
    indShape = 0
    def __init__(self, ObjectToWorld, WorldToObject):
        if (not isinstance(ObjectToWorld, Transform) or not isinstance(WorldToObject, Transform)):
            raise ValueError('The parameters oTw and wTo must be both Transform')
        self.oTw = ObjectToWorld
        self.wTo = WorldToObject


def get_intersect_dataset(shape_name, r, t=None, is_intersection=False, u=None, v=None,
                          dpdu=None, dpdv=None, diag_calc=False):
    """
    Create dataset containing the intersection test information

    - The intersect method return (when parameter ds_output = False) of all quadrics, 
      Triangle and TriangleMesh classes gives directly the inputs of this function

    Parameters
    ----------
    shape_name : str
        The shape class name
    r : Ray
        The ray(s) used for the intersection test
    t : float | 1-D ndarray | 2-D ndarray, optional
        The t ray variable for its first intersection at the shape surface
    is_intersection : bool | 1-D ndarray | 2-D ndarray, optional
        If there is an intersection -> True, else False
    u : float | 1-D ndarray | 2-D ndarray, optional
        The u coordinate of the parametric representation
    v : float | 1-D ndarray | 2-D ndarray, optional
        The u coordinate of the parametric representation
    dpdu : 1-D ndarray | 2-D ndarray, optional
        The surface partial derivative of phit with respect to u
    dpdv : 1-D ndarray | 2-D ndarray, optional
        The surface partial derivative of phit with respect to v
    diag_cal : bool, optional
            This indicates whether a diagonal calculation has been performed

    Returns
    -------
    out : xr.Dataset
        Look-up table with the intersection information
    """
    if not isinstance(r, Ray): raise ValueError('The r parameter must be a Ray')
    if dpdu is None: dpdu = np.array([np.nan, np.nan, np.nan])
    if dpdv is None: dpdv = np.array([np.nan, np.nan, np.nan])

    is_r_arr = isinstance(r.o.x, np.ndarray)
    if isinstance(t, np.ndarray):
        is_obj_arr = (len(t.shape) == 2) or (len(t.shape) == 1 and not is_r_arr)
    else:
        is_obj_arr = False

    ds = xr.Dataset(coords={'xyz':np.arange(3)})
    not_int = np.logical_not(is_intersection)
    
    if is_r_arr:
        nrays = len(r.o.x)
        ds['o'] = xr.DataArray(r.o.to_numpy(), dims=['nrays', 'xyz'])
        ds['d'] = xr.DataArray(r.d.to_numpy(), dims=['nrays', 'xyz'])
        mint = np.zeros(nrays, dtype=np.float64)
        maxt = np.zeros_like(mint)
        mint[:] = r.mint
        maxt[:] = r.maxt
        ds['mint'] = xr.DataArray(mint, dims=['nrays'])
        ds['maxt'] = xr.DataArray(maxt, dims=['nrays'])
    else:
        ds['o'] = xr.DataArray(r.o.to_numpy(), dims=['xyz'])
        ds['d'] = xr.DataArray(r.d.to_numpy(), dims=['xyz'])
        ds['mint'] = xr.DataArray(r.mint)
        ds['maxt'] = xr.DataArray(r.maxt)

    if (not isinstance(t, np.ndarray)):
        ds['is_intersection'] = xr.DataArray(is_intersection)
        ds['thit'] = t
        ds['u'] = u
        ds['v'] = v
        if t is None: p = np.array([np.nan, np.nan, np.nan])
        else:
            p = r(t)
            p = p.to_numpy()
        ds['phit'] = xr.DataArray(p, dims='xyz')
        if t is None: n = np.array([np.nan, np.nan, np.nan])
        else:
            n = face_forward(Normal(normalize(cross(Vector(dpdu), Vector(dpdv)))), -r.d)
            n = n.to_numpy()
        ds['nhit'] = xr.DataArray(n, dims='xyz')
        ds['dpdu'] = xr.DataArray(dpdu, dims='xyz')
        ds['dpdv'] = xr.DataArray(dpdv, dims='xyz')
    elif (is_obj_arr and not is_r_arr): # multiple obj and 1 ray
        ds.attrs.update({'nobj': len(t)})
        ds['is_intersection'] = xr.DataArray(is_intersection, dims=['nobj'])
        ds['thit'] = xr.DataArray(t, dims=['nobj'])
        u[not_int] = None
        v[not_int] = None
        ds['u'] = xr.DataArray(u, dims=['nobj'])
        ds['v'] = xr.DataArray(v, dims=['nobj'])
        p = r(t).to_numpy()
        ds['phit'] = xr.DataArray(p, dims=['nobj', 'xyz'])
        n = face_forward(Normal(normalize(cross(Vector(dpdu), Vector(dpdv)))), -r.d)
        n = n.to_numpy()
        n[not_int] = None
        ds['nhit'] = xr.DataArray(n, dims=['nobj', 'xyz'])
        dpdu[not_int] = None
        dpdv[not_int] = None
        ds['dpdu'] = xr.DataArray(dpdu, dims=['nobj', 'xyz'])
        ds['dpdv'] = xr.DataArray(dpdv, dims=['nobj', 'xyz'])
    elif (not is_obj_arr and is_r_arr): # 1 obj and multiple rays
        ds.attrs.update({'nrays': nrays})
        ds['is_intersection'] = xr.DataArray(is_intersection, dims=['nrays'])
        ds['thit'] = xr.DataArray(t, dims=['nrays'])
        u[not_int] = None
        v[not_int] = None
        ds['u'] = xr.DataArray(u, dims=['nrays'])
        ds['v'] = xr.DataArray(v, dims=['nrays'])
        p = r(t).to_numpy()
        ds['phit'] = xr.DataArray(p, dims=['nrays', 'xyz'])
        n = face_forward(Normal(normalize(cross(Vector(dpdu), Vector(dpdv)))), -r.d)
        n = n.to_numpy()
        n[not_int] = None
        ds['nhit'] = xr.DataArray(n, dims=['nrays', 'xyz'])
        dpdu = np.full((nrays, 3), dpdu, dtype=np.float64)
        dpdv = np.full((nrays, 3), dpdv, dtype=np.float64)
        dpdu[not_int] = None
        dpdv[not_int] = None
        ds['dpdu'] = xr.DataArray(dpdu, dims=['nrays', 'xyz'])
        ds['dpdv'] = xr.DataArray(dpdv, dims=['nrays', 'xyz'])
    elif (diag_calc or shape_name == 'TriangleMesh'): # multiple shape and rays but calc only the diagonal
        ndiag = len(t)
        if shape_name == 'TriangleMesh' and not diag_calc:
            dim_name = 'nrays'
            ds.attrs.update({'nrays': ndiag})
        else:
            dim_name = 'ndiag'
            ds.attrs.update({'nobj': ndiag, 'nrays': ndiag, 'ndiag':ndiag})
        ds['is_intersection'] = xr.DataArray(is_intersection, dims=[dim_name])
        ds['thit'] = xr.DataArray(t, dims=[dim_name])
        u[not_int] = None
        v[not_int] = None
        ds['u'] = xr.DataArray(u, dims=[dim_name])
        ds['v'] = xr.DataArray(v, dims=[dim_name])
        p = r(t).to_numpy()
        ds['phit'] = xr.DataArray(p, dims=[dim_name, 'xyz'])
        
        n = face_forward(Normal(normalize(cross(Vector(dpdu), Vector(dpdv)))), -r.d)
        n = n.to_numpy()
        n[not_int] = None
        ds['nhit'] = xr.DataArray(n, dims=[dim_name, 'xyz'])
        dpdu[not_int] = None
        dpdv[not_int] = None
        ds['dpdu'] = xr.DataArray(dpdu, dims=[dim_name, 'xyz'])
        ds['dpdv'] = xr.DataArray(dpdv, dims=[dim_name, 'xyz'])
    else: # multiple shape and rays, 2-D output
        nobj = t.shape[0]
        ds.attrs.update({'nobj': nobj, 'nrays': nrays})
        ds['is_intersection'] = xr.DataArray(is_intersection, dims=['nobj', 'nrays'] )
        ds['thit'] = xr.DataArray(t, dims=['nobj', 'nrays'])
        u[not_int] = None
        v[not_int] = None
        ds['u'] = xr.DataArray(u, dims=['nobj', 'nrays'])
        ds['v'] = xr.DataArray(v, dims=['nobj', 'nrays'])

        n = np.zeros((nobj, nrays,3), dtype=np.float64)
        p = np.zeros_like(n)

        if nobj >= nrays:
            ro = r.o.to_numpy()
            rd = r.d.to_numpy()
            n_bis = Normal(normalize(cross(Vector(dpdu), Vector(dpdv))))
            for ir in range (0, nrays):
                ri = Ray(Point(ro[ir,:]), Vector(rd[ir,:]), mint[ir], maxt[ir])
                n[:,ir,:] = face_forward(n_bis, -ri.d).to_numpy()
                p[:,ir,:] = ri(t[:,ir]).to_numpy()
        else:
            for iobj in range (0, nobj):
                n_bis = Normal(normalize(cross(Vector(dpdu[iobj,:]), Vector(dpdv[iobj,:]))))
                n[iobj,:,:] = face_forward(n_bis, -r.d).to_numpy()
                p[iobj,:,:] = r(t[iobj,:]).to_numpy()
        n[not_int,:] = None
        
        # n = face_forward(Normal(normalize(cross(Vector(dpdu), Vector(dpdv)))), -r.d)
        ds['phit'] = xr.DataArray(p, dims=['nobj', 'nrays', 'xyz'])
        ds['nhit'] = xr.DataArray(n, dims=['nobj', 'nrays', 'xyz'])
        
        dpdu_2d = np.repeat(dpdu[:,np.newaxis,:], nrays, axis=1)
        dpdv_2d = np.repeat(dpdv[:,np.newaxis,:], nrays, axis=1)
        dpdu_2d[not_int,:] = None
        dpdv_2d[not_int,:] = None
        ds['dpdu'] = xr.DataArray(dpdu_2d, dims=['nobj', 'nrays', 'xyz'])
        ds['dpdv'] = xr.DataArray(dpdv_2d, dims=['nobj', 'nrays', 'xyz'])

    ds['o'].attrs = {'type': 'Point', 'description':'the x, y and z components of the ray point'}
    ds['d'].attrs = {'type': 'Vector', 'description':'the x, y and z components of the ray vector'}
    ds['mint'].attrs = {'description':'the mint attribut of the ray'}
    ds['maxt'].attrs = {'description':'the maxt attribut of the ray'}
    ds['is_intersection'].attrs = {'description':'this variable tells if there is an intersection between the ray and the shape'}
    ds['thit'].attrs = {'description':'the t ray factor for the intersection point calculation'}
    ds['u'].attrs = {'description':'the u coordinate of the parametric representation'}
    ds['v'].attrs = {'description':'the v coordinate of the parametric representation'}
    ds['phit'].attrs = {'type': 'Point', 'description':'the x, y and z components of the intersection point'}
    ds['nhit'].attrs = {'type': 'Normal', 'description':'the x, y and z components of the normal at the intersection point'}
    ds['dpdu'].attrs = {'type': 'Vector', 'description':'the surface partial derivative of phit with respect to u'}
    ds['dpdv'].attrs = {'type': 'Vector', 'description':'the surface partial derivative of phit with respect to v'}
    ds.attrs = {'shape': shape_name}
    date = datetime.now().strftime("%Y-%m-%d")  
    ds.attrs.update({'date':date, 'version': VERSION})
    return ds



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.basic import Vector, Point, Normal, Ray, BBox
from geoclide.vecope import normalize
import numpy as np
from numpy.linalg import inv
import math
import warnings


class Transform(object):
    '''
    Represents 3D geometric transformation(s) using a 4x4 matrix or ntx4x4 matrix, 
    where nT is the number of transformations

    It allows translation, rotation and scalling. It can be applied to vectors, points,
    normals and rays

    Parameters
    ----------
    m : Transform | 2-D ndarray | 3-D ndarray, optional
        The matrix of the transformation(s)
    mInv : Transform | 2-D ndarray | 3-D ndarray, optional
        The inverse matrix of the transformation(s)

    Examples
    --------
    >>> import geoclide as gc
    >>> t1 = gc.Transform()
    >>> t1
    m=
    array(
    [[1. 0. 0. 0.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]] )
    mInv=
    array(
    [[1. 0. 0. 0.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]] )
    '''

    def __init__(self, m = None, mInv = None):
        if (isinstance(m, Transform)):
            self.m = m.m
            self.mInv = m.mInv
        elif (m is None and mInv is None):
            self.m = np.identity(4)
            self.mInv = self.m.copy()
        elif (isinstance(m, np.ndarray) and mInv is None):
            if ( (len(m.shape) == 2 and m.shape != (4,4)) or
                 (len(m.shape) == 3 and m.shape[1] != 4 and m.shape[2] != 4)):
                raise ValueError("The m parameter must be an np.array of shape (4,4) or (nT,4,4)")
            self.m = m
            self.mInv = inv(m)
        elif (m is None and isinstance(mInv, np.ndarray)):
            if ( (len(mInv.shape) == 2 and mInv.shape != (4,4)) or
                 (len(mInv.shape) == 3 and mInv.shape[1] != 4 and mInv.shape[2] != 4)):
                raise ValueError("The mInv parameter must be an np.array of shape (4,4) or (nT,4,4)")
            self.m = inv(mInv)
            self.mInv = mInv
        elif (isinstance(m, np.ndarray) and isinstance(mInv, np.ndarray)):
            if ( (len(m.shape) == 2 and m.shape != (4,4)) or
                 (len(m.shape) == 3 and m.shape[1] != 4 and m.shape[2] != 4)):
                raise ValueError("The m parameter must be an np.array of shape (4,4) or (nT,4,4)")
            if ( (len(mInv.shape) == 2 and mInv.shape != (4,4)) or
                 (len(mInv.shape) == 3 and mInv.shape[1] != 4 and mInv.shape[2] != 4)):
                raise ValueError("The mInv parameter must be an np.array of shape (4,4) or (nT,4,4)")
            self.m = m
            self.mInv = mInv
        else:
            raise ValueError("Wrong parameter value(s) for Transform")

    def __eq__(self, t):
        if (not isinstance(t, Transform)):
            raise ValueError("Equality with a Transform must be only with another Transform")
        
        self.m = t.m
        self.mInv = t.mInv
        
    def __mul__(self, t): 
        if (not isinstance(t, Transform)):
            raise ValueError('A transform can be multiplied only by another Transform')
        
        return Transform(self.m@t.m, t.mInv@self.mInv)
    
    def __call__(self, c, diag_calc=False, flatten=False):
        """
        Apply the transformations

        Parameters
        ----------
        c : Vector | Point | Normal | Ray | BBox
            The vector(s)/point(s)/normal(s)/ray(s)/bounding box(es) to which the 
            transformation is applied
        diag_calc : bool, optional
            Perform diagonal calculations between c(i) and tranformation(i). The number of 
            transformations must be equal to the number of vectors/points/ ... 
        
        Returns
        -------
        out : Vector | Point | Normal | Ray | BBox | 1-D array
            The vector(s)/point(s)/normal(s)/ray(s)/bounding box(es) after the 
            application of the transformation(s). In case of several transformations, it returns 
            a 1-D ndarray of dtype equals to the c parameter type, but if flatten is True 
            returns directly an object of same type as the c parameter.

        Examples
        --------
        >>> import geoclide as gc
        >>> t = gc.get_translate_tf(gc.Vector(5., 5., 5.))
        >>> p = gc.Point(0., 0., 0.)
        >>> t[p]
        Point(5.0, 5.0, 5.0)
        """
        if len(self.m.shape) == 3: # Case with several transformations in Transform
            is_vector = isinstance(c, Vector)
            is_point = isinstance(c, Point)
            is_normal = isinstance(c, Normal)
            if is_vector or is_point or is_normal:
                nT = self.m.shape[0]
                if is_vector or is_point: mat = np.moveaxis(self.m, 0,2)
                else: mat = np.moveaxis(self.mInv, 0,2) # if is_normal
                is_c_arr = isinstance(c.x, np.ndarray)
                key_bis = np.arange(nT)
                if flatten: use_flatten = False
                if is_c_arr and not diag_calc:
                    mat = mat[:,:,np.newaxis,:]
                    x = c.x[:,np.newaxis]
                    y = c.y[:,np.newaxis]
                    z = c.z[:,np.newaxis]
                    if flatten :
                        keys = (slice(None), key_bis)
                        use_flatten = True
                    else : 
                        keys = [(slice(None), k) for k in key_bis]
                else: # if diag_calc = True or if not is_c_arr
                    x = c.x
                    y = c.y
                    z = c.z
                    keys = key_bis
            if is_vector:
                xv = mat[0,0]*x + mat[0,1]*y + mat[0,2]*z
                yv = mat[1,0]*x + mat[1,1]*y + mat[1,2]*z
                zv = mat[2,0]*x + mat[2,1]*y + mat[2,2]*z
                if flatten:
                    if use_flatten :
                        vectors = Vector(xv[keys].flatten('F'), yv[keys].flatten('F'), zv[keys].flatten('F'))
                    else :
                        vectors = Vector(xv[keys], yv[keys], zv[keys])
                else:
                    vectors = np.empty(nT, dtype=Vector)
                    for iv in range (0, nT):
                        vectors[iv] = Vector(xv[keys[iv]], yv[keys[iv]], zv[keys[iv]])
                return vectors
            elif is_point:
                xp = mat[0,0]*x + mat[0,1]*y + mat[0,2]*z + mat[0,3]
                yp = mat[1,0]*x + mat[1,1]*y + mat[1,2]*z + mat[1,3]
                zp = mat[2,0]*x + mat[2,1]*y + mat[2,2]*z + mat[2,3]
                wp = mat[3,0]*x + mat[3,1]*y + mat[3,2]*z + mat[3,3]
                if flatten:
                    if use_flatten:
                        points = Point(xp[keys].flatten('F'), yp[keys].flatten('F'), zp[keys].flatten('F'))
                        points /= wp[keys].flatten('F')
                    else:
                        points = Point(xp[keys], yp[keys], zp[keys])/wp[keys]
                else:
                    points = np.empty(nT, dtype=Point)
                    for ip in range (0, nT):
                        if ((not isinstance(wp[keys[ip]], np.ndarray) and wp[keys[ip]] == 1) or
                            (isinstance(wp[keys[ip]], np.ndarray) and np.all(wp[keys[ip]] == 1)) ):
                            points[ip] = Point(xp[keys[ip]], yp[keys[ip]], zp[keys[ip]])
                        else: 
                            points[ip] = Point(xp[keys[ip]], yp[keys[ip]], zp[keys[ip]])/wp[keys[ip]]
                return points
            elif is_normal:
                xn = mat[0,0]*x + mat[1,0]*y + mat[2,0]*z
                yn = mat[0,1]*x + mat[1,1]*y + mat[2,1]*z
                zn = mat[0,2]*x + mat[1,2]*y + mat[2,2]*z
                if flatten:
                    if use_flatten:
                        normals = Normal(xn[keys].flatten('F'), yn[keys].flatten('F'), zn[keys].flatten('F'))
                    else:
                        normals = Normal(xn[keys], yn[keys], zn[keys])
                else:
                    normals = np.empty(nT, dtype=Vector)
                    for inorm in range (0, nT):
                        normals[inorm] = Normal(xn[keys[inorm]], yn[keys[inorm]], zn[keys[inorm]])
                return normals
            elif isinstance(c, Ray):
                origins = self(c.o, diag_calc, flatten)
                directions = self(c.d, diag_calc, flatten)
                if flatten:
                    rays = Ray(origins, directions, mint=c.mint, maxt=c.maxt)
                else:
                    nT = self.m.shape[0]
                    rays = np.empty(nT, dtype=Ray)
                    for ir in range (0, nT):
                        rays[ir] = Ray(origins[ir], directions[ir], mint=c.mint, maxt=c.maxt)
                return rays
            elif isinstance(c, BBox):
                p0 = self(c.p0, diag_calc, flatten)
                v0 = self(c.p1-c.p0, diag_calc, flatten)
                v1 = self(c.p3-c.p0, diag_calc, flatten)
                v2 = self(c.p4-c.p0, diag_calc, flatten)
                if flatten:
                    b = BBox()
                    b = b.union(p0)
                    b = b.union(p0+v0)
                    b = b.union(p0+(v0+v1))
                    b = b.union(p0+v1)
                    b = b.union(p0+v2)
                    b = b.union(p0+(v0+v2))
                    b = b.union(p0+(v0+v1+v2))
                    b = b.union(p0+(v1+v2))
                    bboxes = b
                else:
                    nT = self.m.shape[0]
                    bboxes = np.empty(nT, dtype=BBox)
                    for ib in range (0, nT):
                        b = BBox()
                        b = b.union(p0[ib])
                        b = b.union(p0[ib]+v0[ib])
                        b = b.union(p0[ib]+(v0[ib]+v1[ib]))
                        b = b.union(p0[ib]+v1[ib])
                        b = b.union(p0[ib]+v2[ib])
                        b = b.union(p0[ib]+(v0[ib]+v2[ib]))
                        b = b.union(p0[ib]+(v0[ib]+v1[ib]+v2[ib]))
                        b = b.union(p0[ib]+(v1[ib]+v2[ib]))
                        bboxes[ib] = b
                return bboxes
            else:
                raise ValueError('Unknown type for transformations')
        else:
            if isinstance(c, Vector):
                xv = self.m[0,0]*c.x + self.m[0,1]*c.y + self.m[0,2]*c.z
                yv = self.m[1,0]*c.x + self.m[1,1]*c.y + self.m[1,2]*c.z
                zv = self.m[2,0]*c.x + self.m[2,1]*c.y + self.m[2,2]*c.z
                return Vector(xv, yv, zv)
            elif isinstance(c, Point):
                xp = self.m[0,0]*c.x + self.m[0,1]*c.y + self.m[0,2]*c.z + self.m[0,3]
                yp = self.m[1,0]*c.x + self.m[1,1]*c.y + self.m[1,2]*c.z + self.m[1,3]
                zp = self.m[2,0]*c.x + self.m[2,1]*c.y + self.m[2,2]*c.z + self.m[2,3]
                wp = self.m[3,0]*c.x + self.m[3,1]*c.y + self.m[3,2]*c.z + self.m[3,3]
                if ((not isinstance(wp, np.ndarray) and wp == 1) or
                    (isinstance(wp, np.ndarray) and np.all(wp == 1)) ):
                    return Point(xp, yp, zp)
                else: 
                    return Point(xp, yp, zp)/wp
            elif isinstance(c, Normal):
                xn = self.mInv[0,0]*c.x + self.mInv[1,0]*c.y + self.mInv[2,0]*c.z
                yn = self.mInv[0,1]*c.x + self.mInv[1,1]*c.y + self.mInv[2,1]*c.z
                zn = self.mInv[0,2]*c.x + self.mInv[1,2]*c.y + self.mInv[2,2]*c.z
                return Normal(xn, yn, zn)
            elif isinstance(c, Ray):
                return Ray(self(c.o), self(c.d), mint=c.mint, maxt=c.maxt)
            elif isinstance(c, BBox):
                b = BBox()
                p0 = self(c.p0)
                v0 = self(c.p1-c.p0)
                v1 = self(c.p3-c.p0)
                v2 = self(c.p4-c.p0)
                b = b.union(p0)
                b = b.union(p0+v0)
                b = b.union(p0+(v0+v1))
                b = b.union(p0+v1)
                b = b.union(p0+v2)
                b = b.union(p0+(v0+v2))
                b = b.union(p0+(v0+v1+v2))
                b = b.union(p0+(v1+v2))
                return b
            else:
                raise ValueError('Unknown type for transformations')
    
    def __getitem__(self, c, diag_calc):
        """
        Apply the transformations

        Parameters
        ----------
        c : Vector | Point | Normal | Ray | BBox
            The Vector/Point/Normal/Ray/BBox to which the transformation is applied
        diag_calc : bool, optional
            Perform diagonal calculations between c(i) and tranformation(i). The number of 
            transformations must be equal to the number of vectors / points / ... 
        
        Returns
        -------
        out : Vector | Point | Normal | Ray | BBox | 1-D array
            The Vector/Point/Normal/Ray/BBox after the transformation, or in case 
            several transformations are given return a 1-D ndarray where dtype is equal 
            to one of the previously mentionned classes.

        Examples
        --------
        >>> import geoclide as gc
        >>> t = gc.get_translate_tf(gc.Vector(5., 5., 5.))
        >>> p = gc.Point(0., 0., 0.)
        >>> t[p]
        Point(5.0, 5.0, 5.0)
        """
        warnings.simplefilter('always', DeprecationWarning)
        warn_message = "\nApplying the transformation through square brackets is deprecated\n" + \
            "as of version 2.1.0 and will be no more possible in the future.\n" + \
            "Please use parenthesis instead."
        warnings.warn(warn_message, DeprecationWarning, stacklevel=1)
        return self(c, diag_calc)

    def __str__(self):
        print("m=\n", self.m, "\nmInv=\n", self.mInv)
        return ""
    
    def __repr__(self):
        print("m=\narray(\n", self.m, ")\nmInv=\narray(\n",self.mInv,")")
        return ""

    def inverse(self):
        """
        Inverse the transformation(s) matrix

        Parameters
        ----------
        t : Transform
            The transformation(s) to be inversed

        Returns
        -------
        out : Transform
            The inversed transformation(s)
        """
        return get_inverse_tf(self)

    def is_identity(self):
        return (self.m[0,0] == 1) and (self.m[0,1] == 0) and (self.m[0,2] == 0) and \
            (self.m[0,3] == 0) and (self.m[1,0] == 0) and (self.m[1,1] == 1) and \
            (self.m[1,2] == 0) and (self.m[1,3] == 0) and (self.m[2,0] == 0) and \
            (self.m[2,1] == 0) and (self.m[2,2] == 1) and (self.m[2,3] == 0) and \
            (self.m[3,0] == 0) and (self.m[3,1] == 0) and (self.m[3,2] == 0) and \
            (self.m[3,3] == 1)

    def translate(self, v):
        """
        Update the self transformation(s) by adding a translate transformation(s)
        
        Parameters
        ----------
        v : Vector
            The vector(s) used for the transformation(s)

        Returns
        -------
        t : Transform
            The product of the self transformation(s) and the translate transformation(s)

        examples
        --------
        >>> import geoclide as gc
        >>> t = Transform()
        >>> t = t.translate(gc.Vector(5.,0.,0.))
        >>> t
        m=
        array(
        [[1. 0. 0. 5.]
        [0. 1. 0. 0.]
        [0. 0. 1. 0.]
        [0. 0. 0. 1.]] )
        mInv=
        array(
        [[ 1.  0.  0. -5.]
        [ 0.  1.  0.  0.]
        [ 0.  0.  1.  0.]
        [ 0.  0.  0.  1.]] )
        """
        t = get_translate_tf(v)
        return self*t

    def scale(self, v):
        """
        Update the self transformation(s) by adding a scale transformation(s)

        Parameters
        ----------
        v : Vector
            The vector(s) used for scale transformation(s)

        Returns
        -------
        t : Transform
            The product of the self transformation(s) and the scale transformation(s) 
            matrices
        """
        t = get_scale_tf(v)
        return self*t

    def rotateX(self, angle):
        """
        Update the self transformation(s) by adding a rotateX transformation(s)

        Parameters
        ----------
        angle : float | 1-D ndarray
            The angle(s) in degrees for the rotation(s) around the x axis

        Returns
        -------
        t : Transform
            The product of the self transformation(s) and the rotateX transformation(s) 
            matrices
        """
        t = get_rotateX_tf(angle)
        return self*t

    def rotateY(self, angle):
        """
        Update the self transformation(s) by adding a rotateY transformation(s)

        Parameters
        ----------
        angle : float | 1-D ndarray
            The angle(s) in degrees for the rotation(s) around the y axis

        Returns
        -------
        t : Transform
            The product of the self transformation(s) and the rotateY transformation(s) 
            matrices
        """
        t = get_rotateY_tf(angle)
        return self*t

    def rotateZ(self, angle):
        """
        Update the self transformation(s) by adding a rotateZ transformation(s)

        Parameters
        ----------
        angle : float | 1-D ndarray
            The angle(s) in degrees for the rotation(s) around the Z axis

        Returns
        -------
        t : Transform
            The product of the initial transformation(s) and the rotateZ transformation(s) 
            matrices
        """
        t = get_rotateZ_tf(angle)
        return self*t

    def rotate(self, angle, axis, diag_calc=False):
        """
        Update the self transformation(s) by adding a rotate transformation(s)

        .. warning::
            The angle parameter can be a 1-D array only if axis parameter is a Vector/Normal 
            with scalar x, y, z components, or if the parameter diag_calc=True

        Parameters
        ----------
        angle : float | 1-D ndarray
            The angle(s) in degrees for the rotation(s)
        axis : Vector | Normal
            The rotation(s) is/are performed around the vector(s)/normal(s) axis/axes
        diag_calc : bool, optional
                Perform diagonal calculations in case angle is a 1-D ndarray and axis is a 
                Vector/Normal with 1-D ndarray x, y, z components. Use angle(i) with axis(i) 
                to calculate transformation(i)

        Returns
        -------
        t : Transform
            The product of the self transformation(s) and the rotate transformation(s) 
            matrices
        """
        t = get_rotate_tf(angle, axis, diag_calc=diag_calc)
        return self*t


def get_inverse_tf(t):
    """
    Get the inverse transformation(s)

    Parameters
    ----------
    t : Transform
        The transformation(s) to be inversed

    Returns
    -------
    out : Transform
        The inversed transformation(s)
    """
    return Transform(t.mInv, t.m)
    

def get_translate_tf(v):
    """
    Get the translate transformation(s)

    Parameters
    ----------
    v : Vector
        The vector(s) used for the translate transformation(s)

    Returns
    -------
    t : Transform
        The translate transformation(s)

    examples
    --------
    >>> import geoclide as gc
    >>> t = gc.get_translate_tf(gc.Vector(5.,0.,0.))
    >>> t
    m=
    array(
    [[1. 0. 0. 5.]
    [0. 1. 0. 0.]
    [0. 0. 1. 0.]
    [0. 0. 0. 1.]] )
    mInv=
    array(
    [[ 1.  0.  0. -5.]
    [ 0.  1.  0. -0.]
    [ 0.  0.  1. -0.]
    [ 0.  0.  0.  1.]] )
    """
    if (not isinstance(v, Vector)):
        raise ValueError("The parameter v must be a Vector")
    if isinstance(v.x, np.ndarray):
        nc = len(v.x)
        m = np.tile(np.identity(4, dtype=np.float64), (nc,1)).reshape(nc,4,4)
        mInv = m.copy()
        m[:,0,3] = v.x
        m[:,1,3] = v.y
        m[:,2,3] = v.z
        mInv[:,0,3] = -v.x
        mInv[:,1,3] = -v.y
        mInv[:,2,3] = -v.z
    else:
        m = np.identity(4)
        mInv = m.copy()
        m[0,3] = v.x
        m[1,3] = v.y
        m[2,3] = v.z
        mInv[0,3] = -v.x
        mInv[1,3] = -v.y
        mInv[2,3] = -v.z
    return Transform(m, mInv)


def get_scale_tf(v):
    """
    Get the scale transformation(s)

    Parameters
    ----------
    v : Vector
        The vector(s) used for scale transformation(s)

    Returns
    -------
    t : Transform
        The scale transformation(s)
    """
    if (not isinstance(v, Vector)):
        raise ValueError("The parameter v must be a Vector")
    
    if isinstance(v.x, np.ndarray):
        nc = len(v.x)
        m = np.tile(np.identity(4, dtype=np.float64), (nc,1)).reshape(nc,4,4)
        mInv = m.copy()
        m[:,0,0] = v.x
        m[:,1,1] = v.y
        m[:,2,2] = v.z
        mInv[:,0,0] = 1./v.x
        mInv[:,1,1] = 1./v.y
        mInv[:,2,2] = 1./v.z
    else:
        m = np.identity(4)
        mInv = m.copy()
        m[0,0] = v.x
        m[1,1] = v.y
        m[2,2] = v.z
        mInv[0,0] = 1./v.x
        mInv[1,1] = 1./v.y
        mInv[2,2] = 1./v.z
    return Transform(m, mInv)


def get_rotateX_tf(angle):
    """
    Get the rotateX transformation(s)

    Parameters
    ----------
    angle : float | 1-D ndarray
        The angle(s) in degrees for the rotation(s) around the x axis

    Returns
    -------
    t : Transform
        The rotateX transformation(s)
    """
    is_ang_arr = isinstance(angle, np.ndarray)

    if (is_ang_arr):
        sin_t = np.sin(angle*(math.pi / 180.))
        cos_t = np.cos(angle*(math.pi / 180.))
        nc = len(angle)
        m = np.tile(np.identity(4, dtype=np.float64), (nc,1)).reshape(nc,4,4)
        m[:,1,1] = cos_t
        m[:,1,2] = -1.*sin_t
        m[:,2,1] = sin_t
        m[:,2,2] = cos_t
        return Transform(m, np.transpose(m, axes=(0,2,1)))
    else:
        sin_t = math.sin(angle*(math.pi / 180.))
        cos_t = math.cos(angle*(math.pi / 180.))
        m = np.identity(4)
        m[1,1] = cos_t
        m[1,2] = -1.*sin_t
        m[2,1] = sin_t
        m[2,2] = cos_t
        return Transform(m, np.transpose(m))


def get_rotateY_tf(angle):
    """
    Get the rotateY transformation(s)

    Parameters
    ----------
    angle : float | 1-D ndarray
        The angle(s) in degrees for the rotation(s) around the y axis

    Returns
    -------
    t : Transform
        The rotateY transformation(s)
    """
    is_ang_arr = isinstance(angle, np.ndarray)

    if (is_ang_arr):
        sin_t = np.sin(angle*(math.pi / 180.))
        cos_t = np.cos(angle*(math.pi / 180.))
        nc = len(angle)
        m = np.tile(np.identity(4, dtype=np.float64), (nc,1)).reshape(nc,4,4)
        m[:,0,0] = cos_t
        m[:,2,0] = -1.*sin_t
        m[:,0,2] = sin_t
        m[:,2,2] = cos_t
        return Transform(m, np.transpose(m, axes=(0,2,1)))
    else:
        sin_t = math.sin(angle*(math.pi / 180.))
        cos_t = math.cos(angle*(math.pi / 180.))
        m = np.identity(4)
        m[0,0] = cos_t
        m[2,0] = -1.*sin_t
        m[0,2] = sin_t
        m[2,2] = cos_t
        return Transform(m, np.transpose(m))


def get_rotateZ_tf(angle):
    """
    Get the rotateZ transformation(s)

    Parameters
    ----------
    angle : float | 1-D ndarray
        The angle(s) in degrees for the rotation(s) around the Z axis

    Returns
    -------
    t : Transform
        The rotateZ transformation(s)
    """
    is_ang_arr = isinstance(angle, np.ndarray)

    if (is_ang_arr):
        sin_t = np.sin(angle*(math.pi / 180.))
        cos_t = np.cos(angle*(math.pi / 180.))
        nc = len(angle)
        m = np.tile(np.identity(4, dtype=np.float64), (nc,1)).reshape(nc,4,4)
        m[:,0,0] = cos_t
        m[:,0,1] = -sin_t
        m[:,1,0] = sin_t
        m[:,1,1] = cos_t
        return Transform(m, np.transpose(m, axes=(0,2,1)))
    else:
        sin_t = math.sin(angle*(math.pi / 180.))
        cos_t = math.cos(angle*(math.pi / 180.))
        m = np.identity(4)
        m[0,0] = cos_t
        m[0,1] = -sin_t
        m[1,0] = sin_t
        m[1,1] = cos_t
        return Transform(m, np.transpose(m))


def get_rotate_tf(angle, axis, diag_calc=False):
    """
    Get the rotate transformation(s) around a given axis/axes

    .. warning::
            The angle parameter can be a 1-D array only if axis parameter is a Vector/Normal 
            with scalar x, y, z components, or if the parameter diag_calc=True

    Parameters
    ----------
    angle : float | 1-D ndarray
        The angle(s) in degrees for the rotation(s)
    axis : Vector | Normal
        The rotation(s) is/are performed around the vector(s)/normal(s) axis/axes
    diag_calc : bool, optional
            Perform diagonal calculations in case angle is a 1-D ndarray and axis is a 
            Vector/Normal with 1-D ndarray x, y, z components. Use angle(i) with axis(i) 
            to calculate transformation(i)

    Returns
    -------
    t : Transform
        The rotate transformation(s)
    """
    if ( (not isinstance(axis, Vector)) and
         (not isinstance(axis, Normal)) ):
        raise ValueError("The parameter axis must be a Vector or a Normal")
    
    is_ang_arr = isinstance(angle, np.ndarray)
    is_axis_arr = isinstance(axis.x, np.ndarray)

    if is_ang_arr and is_axis_arr and not diag_calc:
        raise ValueError("1-D array for angle parameter is allowed only if axis parameter" + \
                         " is a Vector/Normal with scalar components, or if diag_calc=True")

    a = Vector(normalize(axis))
    if is_ang_arr or is_axis_arr:
        nc = 1
        if is_ang_arr:
            nc = max(nc, len(angle))
            s = np.sin(angle*(math.pi / 180.))
            c = np.cos(angle*(math.pi / 180.))
        else:
            s = math.sin(angle*(math.pi / 180.))
            c = math.cos(angle*(math.pi / 180.))
        
        if is_axis_arr: nc = max(nc, len(axis.x))
        m = np.tile(np.identity(4, dtype=np.float64), (nc,1)).reshape(nc,4,4)

        m[:,0,0] = a.x*a.x+(1-a.x*a.x)*c
        m[:,0,1] = a.x*a.y*(1-c)-a.z*s
        m[:,0,2] = a.x*a.z*(1-c)+a.y*s

        m[:,1,0] = a.x*a.y*(1-c)+a.z*s
        m[:,1,1] = a.y*a.y+(1-a.y*a.y)*c
        m[:,1,2] = a.y*a.z*(1-c)-a.x*s

        m[:,2,0] = a.x*a.z*(1-c)-a.y*s
        m[:,2,1] = a.y*a.z*(1-c)+a.x*s
        m[:,2,2] = a.z*a.z+(1-a.z*a.z)*c
        return Transform(m, np.transpose(m, axes=(0,2,1)))
    else:
        s = math.sin(angle*(math.pi / 180.))
        c = math.cos(angle*(math.pi / 180.))
        m = np.identity(4)

        m[0,0] = a.x*a.x+(1-a.x*a.x)*c
        m[0,1] = a.x*a.y*(1-c)-a.z*s
        m[0,2] = a.x*a.z*(1-c)+a.y*s

        m[1,0] = a.x*a.y*(1-c)+a.z*s
        m[1,1] = a.y*a.y+(1-a.y*a.y)*c
        m[1,2] = a.y*a.z*(1-c)-a.x*s

        m[2,0] = a.x*a.z*(1-c)-a.y*s
        m[2,1] = a.y*a.z*(1-c)+a.x*s
        m[2,2] = a.z*a.z+(1-a.z*a.z)*c
        return Transform(m, np.transpose(m))


